/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2007 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Simple central-difference snGrad scheme with quadratic fit correction from 
    a larger stencil.

\*---------------------------------------------------------------------------*/

#include "quadraticFitSnGrad.H"
#include "quadraticFitSnGradData.H"
#include "volFields.H"
#include "surfaceFields.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace fv
{

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

template<class Type>
quadraticFitSnGrad<Type>::~quadraticFitSnGrad()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<class Type>
tmp<GeometricField<Type, fvsPatchField, surfaceMesh> >
quadraticFitSnGrad<Type>::correction
(
    const GeometricField<Type, fvPatchField, volMesh>& vf
) const
{
    typedef typename outerProduct<scalar, Type>::type GradType;

    const fvMesh& mesh = this->mesh();

    const quadraticFitSnGradData& qfd = quadraticFitSnGradData::New
    (
        mesh, 
        centralWeight_
    );

    const labelListList& stCells = qfd.stencilCells();
    const labelListList& stBounds = qfd.stencilBounds();
    const List<scalarList>& f = qfd.fit();

    // construct GeometricField<Type, fvsPatchField, surfaceMesh>
    tmp<GeometricField<Type, fvsPatchField, surfaceMesh> > tssf
    (
        new GeometricField<Type, fvsPatchField, surfaceMesh>
        (
            IOobject
            (
                "snGradQuadCorr("+vf.name()+')',
                vf.instance(),
                mesh,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            mesh,
            dimensioned<GradType>
            (
                "zero",
                vf.dimensions()/dimLength,
                pTraits<GradType>::zero
            )
        )
    );
    
    GeometricField<Type, fvsPatchField, surfaceMesh>& sf = tssf();

    for(label faci = 0; faci < mesh.nInternalFaces(); faci++)
    {
        // Use the fit coefficients to calculate the normal gradient from
        // the values of the field at each cell in stencil

        for(label ip = 0; ip < stCells[faci].size(); ip++)
        {
            sf[faci] += f[faci][ip] * vf[stCells[faci][ip]];
        }
        for(label ib = 0; ib < stBounds[faci].size(); ib++)
        {
            label ip = ib + stCells[faci].size();
            label facei = stBounds[faci][ib];
            label patchi = mesh.boundaryMesh().whichPatch(facei);
            label patchFacei = facei - mesh.boundaryMesh()[patchi].start();

            sf[faci] += f[faci][ip]
                     * vf.boundaryField()[patchi][patchFacei];
        }
    }
    
    return tssf;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace fv

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
